// @todo -  #ifdef VISUALISATION
#include "flamegpu/visualiser/color/ViridisInterpolation.h"

#include <sstream>

namespace flamegpu {
namespace visualiser {

ViridisInterpolation::ViridisInterpolation(const std::string& _variable_name, const float& _min_bound, const float& _max_bound)
    : min_bound(_min_bound)
    , max_bound(_max_bound)
    , variable_name(_variable_name)
    , invert_palette(false) {
    if (min_bound > max_bound) {
        invert_palette = true;
        std::swap(min_bound, max_bound);
    }
}
ViridisInterpolation& ViridisInterpolation::setBounds(const float& _min_bound, const float& _max_bound) {
    min_bound = _min_bound;
    max_bound = _max_bound;
    if (min_bound > max_bound) {
        invert_palette = true;
        std::swap(min_bound, max_bound);
    }
    return *this;
}

std::string ViridisInterpolation::getSrc() const {
    static const std::array<const Color, 256> &raw_colors = rawColors();
    std::stringstream ss;
    ss << "uniform samplerBuffer color_arg;" << "\n";
    // Create a function that maps an index to a color
    // A switch should work, unsure if I can declare an actual array in shaders without using buffers
    ss << "vec3 getRawColor(int i) {" << "\n";
    ss << "    switch (i) {" << "\n";
    for (size_t i = 0; i < raw_colors.size(); ++i) {
        ss << "      case " << i << ":" << "\n";
        ss << "        return vec3(" << raw_colors[i].r << ", " << raw_colors[i].g << ", " << raw_colors[i].b << ");" << "\n";
    }
    ss << "      default: // This should never be hit" << "\n";
    ss << "        return vec3(1, 1, 1);" << "\n";
    ss << "    }" << "\n";
    ss << "}" << "\n";
    ss << "vec4 calculateColor() {" << "\n";
    // Fetch the modifier from texture cache
    ss << "    float modifier = texelFetch(color_arg, gl_InstanceID).x;" << "\n";
    // Clamp the modifier to bounds
    ss << "    modifier = clamp(modifier, float(" << min_bound << "), float(" << max_bound << "));" << "\n";
    // Scale modifier to range [0.0, 255.0]
    ss << "    modifier = 255 * (modifier - " << min_bound << ") / float(" << (max_bound - min_bound) << ");" << "\n";
    // Decide which colors to pull, and how to lerp
    ss << "    const float a = floor(modifier);" << "\n";
    ss << "    const float t = modifier - a;" << "\n";
    // Optionally invert the palette
    ss << "    const vec3 c0 = getRawColor(" << (invert_palette ? "255 - " : "") << "int(a));" << "\n";
    ss << "    const vec3 c1 = getRawColor(" << (invert_palette ? "255 - " : "") << "int(ceil(modifier)));" << "\n";
    // Lerp and return the chosen color (alpha = 1.0f)
    ss << "    return vec4(c0 * (1.0 - t) + c1 * t, 1.0);" << "\n";
    ss << "}" << "\n";
    return ss.str();
}
std::string ViridisInterpolation::getSamplerName() const {
    return "color_arg";
}
std::string ViridisInterpolation::getAgentVariableName() const {
    return variable_name;
}
std::type_index ViridisInterpolation::getAgentVariableRequiredType() const {
    return std::type_index(typeid(float));
}
const std::array<const Color, 256> &ViridisInterpolation::rawColors() {
    static const std::array<const Color, 256> raw_colors = {
        Color{ 0.267004, 0.004874, 0.329415 },
        Color{ 0.268510, 0.009605, 0.335427 },
        Color{ 0.269944, 0.014625, 0.341379 },
        Color{ 0.271305, 0.019942, 0.347269 },
        Color{ 0.272594, 0.025563, 0.353093 },
        Color{ 0.273809, 0.031497, 0.358853 },
        Color{ 0.274952, 0.037752, 0.364543 },
        Color{ 0.276022, 0.044167, 0.370164 },
        Color{ 0.277018, 0.050344, 0.375715 },
        Color{ 0.277941, 0.056324, 0.381191 },
        Color{ 0.278791, 0.062145, 0.386592 },
        Color{ 0.279566, 0.067836, 0.391917 },
        Color{ 0.280267, 0.073417, 0.397163 },
        Color{ 0.280894, 0.078907, 0.402329 },
        Color{ 0.281446, 0.084320, 0.407414 },
        Color{ 0.281924, 0.089666, 0.412415 },
        Color{ 0.282327, 0.094955, 0.417331 },
        Color{ 0.282656, 0.100196, 0.422160 },
        Color{ 0.282910, 0.105393, 0.426902 },
        Color{ 0.283091, 0.110553, 0.431554 },
        Color{ 0.283197, 0.115680, 0.436115 },
        Color{ 0.283229, 0.120777, 0.440584 },
        Color{ 0.283187, 0.125848, 0.444960 },
        Color{ 0.283072, 0.130895, 0.449241 },
        Color{ 0.282884, 0.135920, 0.453427 },
        Color{ 0.282623, 0.140926, 0.457517 },
        Color{ 0.282290, 0.145912, 0.461510 },
        Color{ 0.281887, 0.150881, 0.465405 },
        Color{ 0.281412, 0.155834, 0.469201 },
        Color{ 0.280868, 0.160771, 0.472899 },
        Color{ 0.280255, 0.165693, 0.476498 },
        Color{ 0.279574, 0.170599, 0.479997 },
        Color{ 0.278826, 0.175490, 0.483397 },
        Color{ 0.278012, 0.180367, 0.486697 },
        Color{ 0.277134, 0.185228, 0.489898 },
        Color{ 0.276194, 0.190074, 0.493001 },
        Color{ 0.275191, 0.194905, 0.496005 },
        Color{ 0.274128, 0.199721, 0.498911 },
        Color{ 0.273006, 0.204520, 0.501721 },
        Color{ 0.271828, 0.209303, 0.504434 },
        Color{ 0.270595, 0.214069, 0.507052 },
        Color{ 0.269308, 0.218818, 0.509577 },
        Color{ 0.267968, 0.223549, 0.512008 },
        Color{ 0.266580, 0.228262, 0.514349 },
        Color{ 0.265145, 0.232956, 0.516599 },
        Color{ 0.263663, 0.237631, 0.518762 },
        Color{ 0.262138, 0.242286, 0.520837 },
        Color{ 0.260571, 0.246922, 0.522828 },
        Color{ 0.258965, 0.251537, 0.524736 },
        Color{ 0.257322, 0.256130, 0.526563 },
        Color{ 0.255645, 0.260703, 0.528312 },
        Color{ 0.253935, 0.265254, 0.529983 },
        Color{ 0.252194, 0.269783, 0.531579 },
        Color{ 0.250425, 0.274290, 0.533103 },
        Color{ 0.248629, 0.278775, 0.534556 },
        Color{ 0.246811, 0.283237, 0.535941 },
        Color{ 0.244972, 0.287675, 0.537260 },
        Color{ 0.243113, 0.292092, 0.538516 },
        Color{ 0.241237, 0.296485, 0.539709 },
        Color{ 0.239346, 0.300855, 0.540844 },
        Color{ 0.237441, 0.305202, 0.541921 },
        Color{ 0.235526, 0.309527, 0.542944 },
        Color{ 0.233603, 0.313828, 0.543914 },
        Color{ 0.231674, 0.318106, 0.544834 },
        Color{ 0.229739, 0.322361, 0.545706 },
        Color{ 0.227802, 0.326594, 0.546532 },
        Color{ 0.225863, 0.330805, 0.547314 },
        Color{ 0.223925, 0.334994, 0.548053 },
        Color{ 0.221989, 0.339161, 0.548752 },
        Color{ 0.220057, 0.343307, 0.549413 },
        Color{ 0.218130, 0.347432, 0.550038 },
        Color{ 0.216210, 0.351535, 0.550627 },
        Color{ 0.214298, 0.355619, 0.551184 },
        Color{ 0.212395, 0.359683, 0.551710 },
        Color{ 0.210503, 0.363727, 0.552206 },
        Color{ 0.208623, 0.367752, 0.552675 },
        Color{ 0.206756, 0.371758, 0.553117 },
        Color{ 0.204903, 0.375746, 0.553533 },
        Color{ 0.203063, 0.379716, 0.553925 },
        Color{ 0.201239, 0.383670, 0.554294 },
        Color{ 0.199430, 0.387607, 0.554642 },
        Color{ 0.197636, 0.391528, 0.554969 },
        Color{ 0.195860, 0.395433, 0.555276 },
        Color{ 0.194100, 0.399323, 0.555565 },
        Color{ 0.192357, 0.403199, 0.555836 },
        Color{ 0.190631, 0.407061, 0.556089 },
        Color{ 0.188923, 0.410910, 0.556326 },
        Color{ 0.187231, 0.414746, 0.556547 },
        Color{ 0.185556, 0.418570, 0.556753 },
        Color{ 0.183898, 0.422383, 0.556944 },
        Color{ 0.182256, 0.426184, 0.557120 },
        Color{ 0.180629, 0.429975, 0.557282 },
        Color{ 0.179019, 0.433756, 0.557430 },
        Color{ 0.177423, 0.437527, 0.557565 },
        Color{ 0.175841, 0.441290, 0.557685 },
        Color{ 0.174274, 0.445044, 0.557792 },
        Color{ 0.172719, 0.448791, 0.557885 },
        Color{ 0.171176, 0.452530, 0.557965 },
        Color{ 0.169646, 0.456262, 0.558030 },
        Color{ 0.168126, 0.459988, 0.558082 },
        Color{ 0.166617, 0.463708, 0.558119 },
        Color{ 0.165117, 0.467423, 0.558141 },
        Color{ 0.163625, 0.471133, 0.558148 },
        Color{ 0.162142, 0.474838, 0.558140 },
        Color{ 0.160665, 0.478540, 0.558115 },
        Color{ 0.159194, 0.482237, 0.558073 },
        Color{ 0.157729, 0.485932, 0.558013 },
        Color{ 0.156270, 0.489624, 0.557936 },
        Color{ 0.154815, 0.493313, 0.557840 },
        Color{ 0.153364, 0.497000, 0.557724 },
        Color{ 0.151918, 0.500685, 0.557587 },
        Color{ 0.150476, 0.504369, 0.557430 },
        Color{ 0.149039, 0.508051, 0.557250 },
        Color{ 0.147607, 0.511733, 0.557049 },
        Color{ 0.146180, 0.515413, 0.556823 },
        Color{ 0.144759, 0.519093, 0.556572 },
        Color{ 0.143343, 0.522773, 0.556295 },
        Color{ 0.141935, 0.526453, 0.555991 },
        Color{ 0.140536, 0.530132, 0.555659 },
        Color{ 0.139147, 0.533812, 0.555298 },
        Color{ 0.137770, 0.537492, 0.554906 },
        Color{ 0.136408, 0.541173, 0.554483 },
        Color{ 0.135066, 0.544853, 0.554029 },
        Color{ 0.133743, 0.548535, 0.553541 },
        Color{ 0.132444, 0.552216, 0.553018 },
        Color{ 0.131172, 0.555899, 0.552459 },
        Color{ 0.129933, 0.559582, 0.551864 },
        Color{ 0.128729, 0.563265, 0.551229 },
        Color{ 0.127568, 0.566949, 0.550556 },
        Color{ 0.126453, 0.570633, 0.549841 },
        Color{ 0.125394, 0.574318, 0.549086 },
        Color{ 0.124395, 0.578002, 0.548287 },
        Color{ 0.123463, 0.581687, 0.547445 },
        Color{ 0.122606, 0.585371, 0.546557 },
        Color{ 0.121831, 0.589055, 0.545623 },
        Color{ 0.121148, 0.592739, 0.544641 },
        Color{ 0.120565, 0.596422, 0.543611 },
        Color{ 0.120092, 0.600104, 0.542530 },
        Color{ 0.119738, 0.603785, 0.541400 },
        Color{ 0.119512, 0.607464, 0.540218 },
        Color{ 0.119423, 0.611141, 0.538982 },
        Color{ 0.119483, 0.614817, 0.537692 },
        Color{ 0.119699, 0.618490, 0.536347 },
        Color{ 0.120081, 0.622161, 0.534946 },
        Color{ 0.120638, 0.625828, 0.533488 },
        Color{ 0.121380, 0.629492, 0.531973 },
        Color{ 0.122312, 0.633153, 0.530398 },
        Color{ 0.123444, 0.636809, 0.528763 },
        Color{ 0.124780, 0.640461, 0.527068 },
        Color{ 0.126326, 0.644107, 0.525311 },
        Color{ 0.128087, 0.647749, 0.523491 },
        Color{ 0.130067, 0.651384, 0.521608 },
        Color{ 0.132268, 0.655014, 0.519661 },
        Color{ 0.134692, 0.658636, 0.517649 },
        Color{ 0.137339, 0.662252, 0.515571 },
        Color{ 0.140210, 0.665859, 0.513427 },
        Color{ 0.143303, 0.669459, 0.511215 },
        Color{ 0.146616, 0.673050, 0.508936 },
        Color{ 0.150148, 0.676631, 0.506589 },
        Color{ 0.153894, 0.680203, 0.504172 },
        Color{ 0.157851, 0.683765, 0.501686 },
        Color{ 0.162016, 0.687316, 0.499129 },
        Color{ 0.166383, 0.690856, 0.496502 },
        Color{ 0.170948, 0.694384, 0.493803 },
        Color{ 0.175707, 0.697900, 0.491033 },
        Color{ 0.180653, 0.701402, 0.488189 },
        Color{ 0.185783, 0.704891, 0.485273 },
        Color{ 0.191090, 0.708366, 0.482284 },
        Color{ 0.196571, 0.711827, 0.479221 },
        Color{ 0.202219, 0.715272, 0.476084 },
        Color{ 0.208030, 0.718701, 0.472873 },
        Color{ 0.214000, 0.722114, 0.469588 },
        Color{ 0.220124, 0.725509, 0.466226 },
        Color{ 0.226397, 0.728888, 0.462789 },
        Color{ 0.232815, 0.732247, 0.459277 },
        Color{ 0.239374, 0.735588, 0.455688 },
        Color{ 0.246070, 0.738910, 0.452024 },
        Color{ 0.252899, 0.742211, 0.448284 },
        Color{ 0.259857, 0.745492, 0.444467 },
        Color{ 0.266941, 0.748751, 0.440573 },
        Color{ 0.274149, 0.751988, 0.436601 },
        Color{ 0.281477, 0.755203, 0.432552 },
        Color{ 0.288921, 0.758394, 0.428426 },
        Color{ 0.296479, 0.761561, 0.424223 },
        Color{ 0.304148, 0.764704, 0.419943 },
        Color{ 0.311925, 0.767822, 0.415586 },
        Color{ 0.319809, 0.770914, 0.411152 },
        Color{ 0.327796, 0.773980, 0.406640 },
        Color{ 0.335885, 0.777018, 0.402049 },
        Color{ 0.344074, 0.780029, 0.397381 },
        Color{ 0.352360, 0.783011, 0.392636 },
        Color{ 0.360741, 0.785964, 0.387814 },
        Color{ 0.369214, 0.788888, 0.382914 },
        Color{ 0.377779, 0.791781, 0.377939 },
        Color{ 0.386433, 0.794644, 0.372886 },
        Color{ 0.395174, 0.797475, 0.367757 },
        Color{ 0.404001, 0.800275, 0.362552 },
        Color{ 0.412913, 0.803041, 0.357269 },
        Color{ 0.421908, 0.805774, 0.351910 },
        Color{ 0.430983, 0.808473, 0.346476 },
        Color{ 0.440137, 0.811138, 0.340967 },
        Color{ 0.449368, 0.813768, 0.335384 },
        Color{ 0.458674, 0.816363, 0.329727 },
        Color{ 0.468053, 0.818921, 0.323998 },
        Color{ 0.477504, 0.821444, 0.318195 },
        Color{ 0.487026, 0.823929, 0.312321 },
        Color{ 0.496615, 0.826376, 0.306377 },
        Color{ 0.506271, 0.828786, 0.300362 },
        Color{ 0.515992, 0.831158, 0.294279 },
        Color{ 0.525776, 0.833491, 0.288127 },
        Color{ 0.535621, 0.835785, 0.281908 },
        Color{ 0.545524, 0.838039, 0.275626 },
        Color{ 0.555484, 0.840254, 0.269281 },
        Color{ 0.565498, 0.842430, 0.262877 },
        Color{ 0.575563, 0.844566, 0.256415 },
        Color{ 0.585678, 0.846661, 0.249897 },
        Color{ 0.595839, 0.848717, 0.243329 },
        Color{ 0.606045, 0.850733, 0.236712 },
        Color{ 0.616293, 0.852709, 0.230052 },
        Color{ 0.626579, 0.854645, 0.223353 },
        Color{ 0.636902, 0.856542, 0.216620 },
        Color{ 0.647257, 0.858400, 0.209861 },
        Color{ 0.657642, 0.860219, 0.203082 },
        Color{ 0.668054, 0.861999, 0.196293 },
        Color{ 0.678489, 0.863742, 0.189503 },
        Color{ 0.688944, 0.865448, 0.182725 },
        Color{ 0.699415, 0.867117, 0.175971 },
        Color{ 0.709898, 0.868751, 0.169257 },
        Color{ 0.720391, 0.870350, 0.162603 },
        Color{ 0.730889, 0.871916, 0.156029 },
        Color{ 0.741388, 0.873449, 0.149561 },
        Color{ 0.751884, 0.874951, 0.143228 },
        Color{ 0.762373, 0.876424, 0.137064 },
        Color{ 0.772852, 0.877868, 0.131109 },
        Color{ 0.783315, 0.879285, 0.125405 },
        Color{ 0.793760, 0.880678, 0.120005 },
        Color{ 0.804182, 0.882046, 0.114965 },
        Color{ 0.814576, 0.883393, 0.110347 },
        Color{ 0.824940, 0.884720, 0.106217 },
        Color{ 0.835270, 0.886029, 0.102646 },
        Color{ 0.845561, 0.887322, 0.099702 },
        Color{ 0.855810, 0.888601, 0.097452 },
        Color{ 0.866013, 0.889868, 0.095953 },
        Color{ 0.876168, 0.891125, 0.095250 },
        Color{ 0.886271, 0.892374, 0.095374 },
        Color{ 0.896320, 0.893616, 0.096335 },
        Color{ 0.906311, 0.894855, 0.098125 },
        Color{ 0.916242, 0.896091, 0.100717 },
        Color{ 0.926106, 0.897330, 0.104071 },
        Color{ 0.935904, 0.898570, 0.108131 },
        Color{ 0.945636, 0.899815, 0.112838 },
        Color{ 0.955300, 0.901065, 0.118128 },
        Color{ 0.964894, 0.902323, 0.123941 },
        Color{ 0.974417, 0.903590, 0.130215 },
        Color{ 0.983868, 0.904867, 0.136897 },
        Color{ 0.993248, 0.906157, 0.143936 }
    };
    return raw_colors;
}

}  // namespace visualiser
}  // namespace flamegpu
